#!/usr/bin/env python

# Copyright Amine Namouchi
# snpToolkit is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but without
# any warranty; without even the implied warranty of merchantability or fitness
# for a particular purpose. See the GNU General Public License for more details
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.


__licence__ = 'GPLv3'
__author__ = 'Amine Namouchi'
__author_email__ = 'amine.namouchi@gmail.com'
__version__ = '2.2.3'

import sys
from pathlib import Path
import pysam
from tqdm import tqdm




class CheckCoverage (object):

    def __init__(self, bamfile, location):
        self.bamfile = pysam.AlignmentFile(bamfile)
        self.location = location

    def location_exist(self):
        Listlocation = self.bamfile.references
        if self.location in Listlocation:
            return True
        else:
            return False

    def coverage(self, position):
        for pileup in self.bamfile.pileup(self.location, position, position + 1, truncate=True):
            return pileup.n


def checkBamindex(PathBamFiles):
    noindex = []
    FolderContent = Path(PathBamFiles)
    bamfiles = FolderContent.glob('*.bam')
    for eachFile in bamfiles:
        bam = pysam.AlignmentFile(str(eachFile))
        try:
            bam.check_index()
        except ValueError:
            noindex.append(str(eachFile))
    return [str(path) for path in noindex]


def checksnpToolkitBam(PathBamFiles):
    snpToolkitFolderContent = Path.cwd()
    bamFilterContent = Path(PathBamFiles)
    nameTObam = {}
    for name in snpToolkitFolderContent.glob('*_snpToolkit_SNPs.txt'):
        for bam in list(bamFilterContent.glob('*.bam')):
            if bam.stem in name.stem:
                nameTObam[name.name] = str(bam.parent) + '/' + bam.name
    return nameTObam


def excludeCloseSNPs(list_snps, regions_to_exclude):
    filteredSNPs = []
    Coordinates_to_exclude = []
    Keywords_to_exclude = []
    if 'COORDINATES' in regions_to_exclude.keys():
        Coordinates_to_exclude = [
            int(x) for x in regions_to_exclude['COORDINATES'].split(';')]
    if 'KEYWORDS' in regions_to_exclude.keys():
        Keywords_to_exclude = regions_to_exclude['KEYWORDS'].split(';')
    for snp in list_snps:
        if snp[0] not in Coordinates_to_exclude:
            if '|' in snp[3]:
                if snp[3].split('|')[0] not in Keywords_to_exclude and snp[3].split('|')[1] not in Keywords_to_exclude:
                    filteredSNPs.append(snp)
            else:
                if snp[3] not in Keywords_to_exclude:
                    filteredSNPs.append(snp)
    return filteredSNPs


def snp_distribution(list_snps, PathFiles, regions_to_exclude):

    if len(regions_to_exclude) > 0:
        filteredSNPs = excludeCloseSNPs(list_snps, regions_to_exclude)
    else:
        filteredSNPs = list_snps
    extract_files_content = {}
    for eachFile in PathFiles:
        with open(eachFile, 'r') as f:
            file_content = [[int(l.strip().split('\t')[0])] + l.strip().split(
                '\t')[1:3] + l.strip().split('\t')[8:-1]  for l in f if '##' not in l]
        extract_files_content[eachFile] = file_content

    for i in tqdm(range(len(filteredSNPs)), ascii=True, desc='progress'):
        eachSNP = filteredSNPs[i]
        searchSNP = eachSNP[:]
        for eachFile in extract_files_content.keys():

            if searchSNP in extract_files_content[eachFile]:
                eachSNP.append('1')
            else:
                eachSNP.append('0')

    return filteredSNPs


def snp_distribution_missing(location, ratio, snpType, list_snps, snpToolkitFiles, bamFilter, checkedBam, cutoff, checkRatio, regions_to_exclude):

    if len(regions_to_exclude) > 0:
        filteredSNPs = excludeCloseSNPs(list_snps, regions_to_exclude)
    else:
        filteredSNPs = list_snps

    Reverse_nuc = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}
    communFile = set(checkedBam.keys()).intersection(snpToolkitFiles)
    if len(communFile) > 0:
        FolderContent = Path(bamFilter)
        bamfiles = list(FolderContent.glob('*.bam'))
        for i in tqdm(range(len(filteredSNPs)), ascii=True, desc='progress'):
            eachSNP = filteredSNPs[i]
            searchSNP = eachSNP[:]
            if eachSNP[5] == '+':
                REF = eachSNP[1]
                SNP = eachSNP[2]
            else:
                REF = Reverse_nuc[eachSNP[1]]
                SNP = Reverse_nuc[eachSNP[2]]
            for eachFile in snpToolkitFiles:
                with open(eachFile, 'r') as f:
                    file_content = [[int(l.strip().split('\t')[0])] + l.strip().split(
                        '\t')[1:3] + l.strip().split('\t')[8:-1] for l in f if '##' not in l]
                if searchSNP in file_content:
                    eachSNP.append('1')
                else:
                    if eachFile in list(checkedBam.keys()):
                        bamFile = pysam.AlignmentFile(checkedBam[eachFile])
                        try:
                            depthATposition = 0
                            # bamFile.count_coverage() return four array.arrays of the same length in order A C G T
                            NucleotidesOrder = 'ACGT'
                            NumberOfReads = []
                            k = 0

                            for eachCov in bamFile.count_coverage(location, eachSNP[0] - 1, eachSNP[0]):
                                depthATposition = depthATposition + \
                                    list(eachCov)[0]
                                NumberOfReads.append(
                                    (list(eachCov)[0], NucleotidesOrder[k]))
                                k += 1
                            NumberOfReads.sort(reverse=True)
                            if depthATposition < cutoff:
                                eachSNP.append('?')
                            else:
                                if NumberOfReads[0][1] == SNP:
                                    if NumberOfReads[0][0] / depthATposition >= checkRatio:
                                        eachSNP.append('1')
                                    elif NumberOfReads[0][0] / depthATposition < checkRatio and NumberOfReads[0][0] / depthATposition >= 0.4 :
                                        eachSNP.append('?')
                                    else:
                                        eachSNP.append('0')
                                else:
                                    eachSNP.append('0')

                        except ValueError as e:
                            logging.error(location + ' was not detected in file ' +
                                          checkedBam[eachFile] + '. Please use samtools view -H on this bam file to check the exact name.')
                            sys.exit(0)
                    else:
                        eachSNP.append('0')
    else:
        list_snps_only_bamFiles = []
        for eachFile in list(communFile):
            with open(eachFile, 'r') as f:
                for l in f.readlines():
                    if l[:2] != '##':
                        content = l.strip().split('\t')
                        if float(content[6]) >= float(ratio):
                            snp_coordinate = int(content[0])
                            if snp_coordinate not in list_snps_only_bamFiles:
                                if location in l.split('\t')[-1] or location == l.split('\t')[-1]:
                                    if snpType == 'all':
                                        list_snps_only_bamFiles.append(
                                            snp_coordinate)
                                    elif snpType == 's':
                                        if l.split('\t')[17] == 'Syn':
                                            list_snps_only_bamFiles.append(
                                                snp_coordinate)
                                    elif snpType == 'ns':
                                        if l.split('\t')[17] == 'NS':
                                            list_snps_only_bamFiles.append(
                                                snp_coordinate)
                                    else:
                                        if l.split('\t')[8] == 'intergenic':
                                            list_snps_only_bamFiles.append(
                                                snp_coordinate)

        FolderContent = Path(bamFilter)
        bamfiles = list(FolderContent.glob('*.bam'))
        for i in tqdm(range(len(filteredSNPs)), ascii=True, desc='progress'):
            Flag = False
            if filteredSNPs[i][0] in list_snps_only_bamFiles:
                Flag = True
            eachSNP = filteredSNPs[i]
            searchSNP = eachSNP[:]
            if eachSNP[5] == '+':
                REF = eachSNP[1]
                SNP = eachSNP[2]
            else:
                REF = Reverse_nuc[eachSNP[1]]
                SNP = Reverse_nuc[eachSNP[2]]
            for eachFile in snpToolkitFiles:
                with open(eachFile, 'r') as f:
                    file_content = [[int(l.strip().split('\t')[0])] + l.strip().split(
                        '\t')[1:3] + l.strip().split('\t')[8:-1] for l in f if '##' not in l]
                if searchSNP in file_content:
                    eachSNP.append('1')
                else:
                    if Flag is True:
                        if eachFile in list(checkedBam.keys()):
                            bamFile = pysam.AlignmentFile(checkedBam[eachFile])
                            try:
                                depthATposition = 0
                                # bamFile.count_coverage() return four array.arrays of the same length in order A C G T
                                NucleotidesOrder = 'ACGT'
                                NumberOfReads = []
                                k = 0
                                # minimum quality of count_coverage is 15
                                for eachCov in bamFile.count_coverage(location, eachSNP[0] - 1, eachSNP[0]):
                                    depthATposition = depthATposition + \
                                        list(eachCov)[0]
                                    NumberOfReads.append(
                                        (list(eachCov)[0], NucleotidesOrder[k]))
                                    k += 1
                                NumberOfReads.sort(reverse=True)
                                if depthATposition <= cutoff:
                                    eachSNP.append('?')
                                else:
                                    if NumberOfReads[0][1] == SNP:
                                        if NumberOfReads[0][0] / depthATposition >= 0.8:
                                            eachSNP.append('1')
                                        elif 0.8 > NumberOfReads[0][0] / depthATposition >= 0.4:
                                            eachSNP.append('?')
                                        else:
                                            eachSNP.append('0')
                                    else:
                                        eachSNP.append('0')

                            except ValueError as e:
                                logging.error(location + ' was not detected in file ' +
                                              checkedBam[eachFile] + '. Please use samtools view -H on this bam file to check the exact name.')
                                sys.exit(0)
                        else:
                            eachSNP.append('0')
                    else:
                        eachSNP.append('0')

    return filteredSNPs


def reconstruct_fasta(filehandle, pad, names, list_snps):
    for i in tqdm(range(len(names)), ascii=True, desc='progress'):
        EachSample = names[i]
        sequence = ''
        for eachSNP in list_snps:
            if eachSNP[i + pad] == '1':
                sequence = sequence + eachSNP[2]
            elif eachSNP[i + pad] == '0':
                sequence = sequence + eachSNP[1]
            else:
                sequence = sequence + '?'
        filehandle.write('>' + EachSample + '\n' + sequence + '\n')
        i += 1
    filehandle.close()




#!/usr/bin/env python

# Copyright Amine Namouchi
# snpToolkit is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but without
# any warranty; without even the implied warranty of merchantability or fitness
# for a particular purpose. See the GNU General Public License for more details
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.


__licence__ = 'GPLv3'
__author__ = 'Amine Namouchi'
__author_email__ = 'amine.namouchi@gmail.com'
__version__ = '2.2.3'


import glob
import argparse
import logging
import coloredlogs
from tqdm import tqdm
import pandas as pd
import random
import plotly.graph_objs as go
import plotly.offline as pyo
import plotly.figure_factory as ff
import numpy as np 
import dash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies  import Input, Output
import dash_table as dt
from operator import itemgetter
from subprocess import Popen, PIPE, STDOUT
from annotate_snpToolkit import *
from argsLogger_snpToolkit import *
from plotly.colors import n_colors


color_choice = "aliceblue, antiquewhite, aqua, aquamarine, azure, beige, bisque, black, blanchedalmond, blue, blueviolet, brown, burlywood, cadetblue, chartreuse, chocolate, coral, cornflowerblue,cornsilk, crimson, cyan, darkblue, darkcyan,darkgoldenrod, darkgray, darkgrey, darkgreen,darkkhaki, darkmagenta, darkolivegreen, darkorange,darkorchid, darkred, darksalmon, darkseagreen,darkslateblue, darkslategray, darkslategrey, darkturquoise, darkviolet, deeppink, deepskyblue,dimgray, dimgrey, dodgerblue, firebrick,floralwhite, forestgreen, fuchsia, gainsboro, ghostwhite, gold, goldenrod, gray, grey, green,greenyellow, honeydew, hotpink, indianred, indigo,ivory, khaki, lavender, lavenderblush, lawngreen,lemonchiffon, lightblue, lightcoral, lightcyan,lightgoldenrodyellow, lightgray, lightgrey,lightgreen, lightpink, lightsalmon, lightseagreen,lightskyblue, lightslategray, lightslategrey,lightsteelblue, lightyellow, lime, limegreen,linen, magenta, maroon, mediumaquamarine, mediumblue, mediumorchid, mediumpurple,mediumseagreen, mediumslateblue, mediumspringgreen,mediumturquoise, mediumvioletred, midnightblue, mintcream, mistyrose, moccasin, navajowhite, navy,oldlace, olive, olivedrab, orange, orangered,orchid, palegoldenrod, palegreen, paleturquoise,palevioletred, papayawhip, peachpuff, peru, pink,plum, powderblue, purple, red, rosybrown,royalblue, saddlebrown, salmon, sandybrown,seagreen, seashell, sienna, silver, skyblue,slateblue, slategray, slategrey, snow, springgreen,steelblue, tan, teal, thistle, tomato, turquoise,violet, wheat, white, whitesmoke, yellow,yellowgreen".split(',')


logger = setupLogger()

parser = argparse.ArgumentParser()
parser.add_argument('-i', dest='identifier', required=True, help='provide input file')


args= parser.parse_args()
dna = ('A', 'C', 'T', 'G')
allowed_format = {".vcf", ".vcf.gz", ".vcf.zip"}
FilesToProcess = [FILE for FILE in glob.glob(
    '*' + args.identifier + '*') if FILE.endswith(tuple(allowed_format))]

if len(FilesToProcess) == 0:
    logger.error(
        'No input file detected! Please check your(s) file(s) name(s)...')
    sys.exit(0)
known_origins = {"samtools", "GATK", "freeBayes"}
logger.info('snpToolkit is extracting your data and creating the different plots...')
warnings = []
flag = False
indelFile = False
vcf_data_collection = {}
isolates_collection = []
for i in tqdm(range(len(FilesToProcess)), ascii=True, desc='progress'):
    VcfFile = FilesToProcess[i]
    File_name = VcfFile.split('.vcf')[0]
    isolates_collection.append ({'label':File_name,'value':File_name})
    VcfFile_object = VCFtoolBox(VcfFile)
    vcf_source = VcfFile_object.vcf_generator()
    if vcf_source not in known_origins:
        warnings.append('Are you sure the vcf file ' + File_name +
                        ' was generated using samtools-mpileup, gatk-HaplotypeCaller or freeBayes!'.format(vcf_source))
    else:
        all_variations = VcfFile_object.extract_all_variation()
        indels_inVCF = []
        SNPs_in_VCF = []
        for EachVar in all_variations:
            if EachVar[3] in dna and EachVar[4] in dna:
                SNPs_in_VCF.append(EachVar)
            else:
                indels_inVCF.append(EachVar)
        raw_number_snps = len(SNPs_in_VCF)
        PreProcessedRawSNPs = SNPselect(SNPs_in_VCF, 0, 0, 0)
        extractedRawSNPs = PreProcessedRawSNPs.ExtractSNPinfo(vcf_source)
        vcf_data_collection [File_name] = extractedRawSNPs

diffContigs = []
for each_vcf_File in vcf_data_collection.keys():
    for eachElem in vcf_data_collection[each_vcf_File][0].keys():
        if eachElem not in diffContigs:
            diffContigs.append(eachElem)
diffContigs.sort()
color_picker={}
randNum = list(set(range(1, len(color_choice))))
random.shuffle(randNum)
uniqueRandNum = list(set(randNum))
for x in range (len (diffContigs)):
    color_picker[diffContigs[x]]= uniqueRandNum[x]

data ={}
data2 = {}
all_snps_stats = {}
for each_vcf_File in vcf_data_collection.keys():
    vcf_content = vcf_data_collection[each_vcf_File][0]
    allSNP = []
    allSNP2 = []
    statsSNPs =[]
    if len (vcf_content.keys())>1:
        colors = n_colors('rgb(5, 200, 200)', 'rgb(200, 10, 10)', len(vcf_content.keys()), colortype='rgb')
    else:
        colors = ['rgb(62, 155, 184)']
    snpsSTATS=[]
    for eachElem,color in zip(vcf_content.keys(),colors):
        df_raw = pd.DataFrame(vcf_content[eachElem],[x[0] for x in vcf_content[eachElem]],['Position','REF','SNP','Depth','Depth reference','Depth SNPs','Ratio','Quality'])
        snpsSTATS.append ([eachElem,len(df_raw),len(df_raw[df_raw.Depth >= 3]),len(df_raw[df_raw.Quality >= 20]),len(df_raw[df_raw.Ratio >= 0.9]),len(df_raw[(df_raw.Ratio >= 0.9) & (df_raw.Depth >= 3) & (df_raw.Quality >= 20)])])
        bubble_color = color_picker[eachElem]
        Qrange = [df_raw['Quality'].between(1, 20), df_raw['Quality'].between(21, 50), df_raw['Quality'].between(51, 100), df_raw['Quality'].between(101, 1000000)] 
        values = [5, 10, 15, 20]
        df_raw['Qvalues'] = np.select(Qrange, values)
        allSNP.append(go.Scatter(x=df_raw["Ratio"],y=df_raw["Depth"],mode="markers",name=eachElem,marker=dict(size=df_raw['Qvalues'],color=color,opacity=0.3,showscale=False),showlegend=True))#color=df_raw['Quality']
        allSNP2.append(go.Violin(x=df_raw["Ratio"],name=eachElem,marker=dict(opacity=0.5,color=color),points=False,orientation='h', side='positive',showlegend=True))

    all_snps_stats[each_vcf_File]=pd.DataFrame(columns=['Location', 'Total SNPs','Nb SNP if depth >= 3','Nb SNP if QUAL >= 20','Nb SNP if Ratio >= 0.9','All filters on'],data=snpsSTATS)

    data[each_vcf_File]=allSNP
    data2[each_vcf_File]=allSNP2



app = dash.Dash()

colors={'background':'#111111','text':'#ffffff','textHeader':'#111111'}
app.layout = html.Div(children=[
    html.H2('snpToolkit plots -  VCFs SNPs content',style={'color':colors['textHeader'],'text-align':'center'}),
    dcc.Dropdown(id='isolate',options=isolates_collection,value=isolates_collection[0]['value']),
    html.H3('Number of identified SNPs',style={'color':colors['textHeader']}),
    html.Div(dt.DataTable(columns = [{"name": i, "id": i} for i in ['Location', 'Total SNPs','Nb SNP if depth >= 3','Nb SNP if QUAL >= 20','Nb SNP if Ratio >= 0.9','All filters on']],id='snpTable', style_table={ 'align':'center'},style_cell={'textAlign': 'center', 'border': '1px solid grey'},    style_header={
        'color':'white',
        'backgroundColor': 'rgb(62, 155, 184)',
        'border': '1px solid black'
    })),
    dcc.Graph(id='graph'),
    html.Div(children=[
    dcc.Graph(id='graph2')
    ])])


@app.callback (Output('snpTable','data'),[Input('isolate','value')])
def update_table(selected_isolate):
    return all_snps_stats[selected_isolate].to_dict('rows')

@app.callback (Output('graph','figure'),[Input('isolate','value')])
def update_graph(selected_isolate):
    layout = go.Layout(title='Depth vs Ratio',xaxis={'title':'Ratio'},yaxis={'title':'Depth'},hovermode='closest')
    return {'data':data[selected_isolate],'layout':layout}

@app.callback (Output('graph2','figure'),[Input('isolate','value')])
def update_graph2(selected_isolate):
    layout2 = go.Layout(title='Frequency distribution',xaxis={'title':'Ratio','range':[0,1]}, yaxis={'showticklabels':False}, hovermode=False,autosize=False,height=300,xaxis_showgrid=False, xaxis_zeroline=False)
    return {'data':data2[selected_isolate],'layout':layout2}



if __name__ == '__main__':
    app.run_server()
    




#!/usr/bin/env python

# Copyright Amine Namouchi
# snpToolkit is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but without
# any warranty; without even the implied warranty of merchantability or fitness
# for a particular purpose. See the GNU General Public License for more details
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.


__licence__ = 'GPLv3'
__author__ = 'Amine Namouchi'
__author_email__ = 'amine.namouchi@gmail.com'
__version__ = '2.2.3'


import glob
import ntpath
import sys
import os
import time
import logging
import coloredlogs
from tqdm import tqdm
from itertools import islice
from pathlib import Path
from datetime import datetime
from operator import itemgetter
from subprocess import Popen, PIPE, STDOUT
from argsLogger_snpToolkit import *
from annotate_snpToolkit import *
from combine_snpToolkit import *
from pathlib import Path
import pandas as pd
import pysam

logger = setupLogger()


def annotate(options,VcfFile,annotationDB,snps_output_directory,indels_ouput_directory):
    dna = ('A', 'C', 'T', 'G')
    regions_to_exclude = []
    if options.exclude != None:
        regions_to_exclude = pd.read_csv(options.exclude, sep='\t', header=None).values.tolist()
    known_origins = {"samtools", "GATK", "freeBayes"}


    warnings = []
    flag = False
    indelFile = False
    File_name = VcfFile.split('.vcf')[0]
    VcfFile_object = VCFtoolBox(VcfFile)
    vcf_source = VcfFile_object.vcf_generator()
    if vcf_source not in known_origins:
        warnings.append('Are you sure the vcf file ' + File_name +
                        ' was generated using samtools-mpileup, gatk-HaplotypeCaller or freeBayes!'.format(vcf_source))
    else:
        all_variations = VcfFile_object.extract_all_variation()
        indels_inVCF = []
        SNPs_in_VCF = []
        for EachVar in all_variations:
            if EachVar[3] in dna and EachVar[4] in dna:
                SNPs_in_VCF.append(EachVar)
            else:
                indels_inVCF.append(EachVar)
        raw_number_snps = len(SNPs_in_VCF)

        info1 = '##Total number of SNPs before snpToolkit processing: {}'.format(
            raw_number_snps)

        SNPs_to_process = SNPfiltering(SNPs_in_VCF)

        if options.excludeCloseSNPs != None and len(regions_to_exclude) == 0:
            Fitered_SNPs = SNPs_to_process.FilterCloseSNPs(
                options.excludeCloseSNPs)
            info2 = '##By excluding SNPs that are closer than {} bp to each other, the number of remaining SNPs is: {}'.format(
                options.excludeCloseSNPs, len(Fitered_SNPs))

        elif options.excludeCloseSNPs == None and len(regions_to_exclude) > 0:
            Fitered_SNPs = SNPs_to_process.FilterSNPtoExclude(
                regions_to_exclude)
            info2 = '##By excluding SNPs that are located in the provided regions to exclude, the number of remaining SNPs is: {}'.format(
                len(Fitered_SNPs))

        elif options.excludeCloseSNPs != None and len(regions_to_exclude) > 0:
            Fitered_SNPs = SNPs_to_process.FilterCloseSNPsandInREgions(
                options.excludeCloseSNPs, regions_to_exclude)[1]
            info2 = '##By excluding SNPs that are closer than {} bp to each other, and also located in the provided regions to exclude, the number of remaining SNPs is: {}'.format(
                options.excludeCloseSNPs, len(Fitered_SNPs))

        else:
            Fitered_SNPs = SNPs_in_VCF
            info2 = '##The options -f and -e were not used'

        PreProcessedRawSNPs = SNPselect(
            Fitered_SNPs, 0, 0, 0)
        extractedRawSNPs = PreProcessedRawSNPs.ExtractSNPinfo(vcf_source)

        PreProcessedSNPs = SNPselect(
            Fitered_SNPs, options.quality, options.depth, options.ratio)
        extractedSNPs = PreProcessedSNPs.ExtractSNPinfo(vcf_source)

        if options.ratio == 0.001:
            info3 = '##Filtred SNPs. Among the {} SNPs, the number of those with a quality score >= {}, a depth >= {} and a ratio >= {} is: {}'.format(
                len(Fitered_SNPs), options.quality, options.depth, 0.0, sum([len(x) for x in extractedSNPs[0].values()]))
        else:
            info3 = '##Filtred SNPs. Among the {} SNPs, the number of those with a quality score >= {}, a depth >= {} and a ratio >= {} is: {}'.format(
                len(Fitered_SNPs), options.quality, options.depth, options.ratio, sum([len(x) for x in extractedSNPs[0].values()]))

        genbank_accessions = annotationDB.keys()
        gbank_accessions_order = []
        for eachElem1 in annotationDB['locus_order']:
            for eachElem2 in genbank_accessions:
                if eachElem1 in eachElem2:
                    gbank_accessions_order.append(eachElem2)
        SNPs_To_Map_and_Annotate = SNPannotation(extractedSNPs[0])
        Final_SNP_List = SNPs_To_Map_and_Annotate.mapAndannoSNPs(
            annotationDB)
        ##################################################################
        PreProcessedINDELS = SNPselect(
            indels_inVCF, options.quality, options.depth, options.ratio)
        extractedINDELS = PreProcessedINDELS.ExtractSNPinfo(vcf_source)
        INDELS_To_Map_and_Annotate = SNPannotation(extractedINDELS[0])
        Final_INDELS_List = INDELS_To_Map_and_Annotate.mapAndannoINDELS(
            annotationDB)
        indels_found = False
        for elem in Final_INDELS_List.keys():
            if len(Final_INDELS_List[elem]) > 1:
                indels_found = True
                break
        if indels_found is True:

            outputfile = open(indels_ouput_directory + '/' +
                                File_name + '_snpToolkit_indels.txt', 'w')

            for eachElem in Final_INDELS_List.keys():
                done=[]
                for eachIndel in Final_INDELS_List[eachElem]:
                    if eachIndel[0] not in done:
                        done.append(eachIndel[0])
                        outputfile.write(
                            '\t'.join([str(e) for e in eachIndel]+[eachElem]) + '\n')
        ##################################################################

        info4 = '##After mapping, SNPs were located in: \n##' + \
            '\n##'.join(Final_SNP_List.keys())
        info5 = '##The mapped and annotated SNPs are distributed as follow:'

        summary_columns = [['Location', 'Genes', 'RBS', 'tRNA', 'rRNA', 'ncRNA',
                            'Pseudogenes', 'intergenic', 'Synonymous', 'NonSynonumous']]
        allSNPs = []
        all_data_collection = []
        for eachAccession in genbank_accessions:
            for location in Final_SNP_List.keys():
                if eachAccession == location:
                    data_collection = []
                    snp_distribution = Final_SNP_List[location]
                    genomic_regions = snp_distribution.keys()
                    counts = [location]
                    Synonymous = 0
                    NonSynonumous = 0
                    for eachRegion1 in summary_columns[0][1:]:
                        for eachRegion2 in genomic_regions:
                            if eachRegion1 == eachRegion2:
                                counts.append(
                                    format(len(Final_SNP_List[location][eachRegion1])))
                                for eachSNP in Final_SNP_List[location][eachRegion1]:
                                    eachSNP.append(eachAccession)
                                    data_collection.append(eachSNP)
                                    allSNPs.append(eachSNP)
                                    if eachSNP[-2] == 'Syn':
                                        Synonymous += 1
                                    elif eachSNP[-2] == 'NS':
                                        NonSynonumous += 1
                    counts.append(format(Synonymous))
                    counts.append(format(NonSynonumous))
                    summary_columns.append(counts)
                    all_data_collection.append(data_collection)

        if len(all_data_collection) > 0:
            flag = True
            outputfile = open(snps_output_directory + '/' +
                                File_name + '_snpToolkit_SNPs.txt', 'w')
            infos = ['##snpToolkit=version '+__version__, '##commandline= ' + ' '.join(
                sys.argv[:]), '##VcfFile='+VcfFile, info1, info2, info3, info4, info5]
            for info in infos:
                outputfile.write(info + '\n')
            outputfile.write('##'+'\t'.join(summary_columns[0]) + '\n')
            for info in summary_columns[1:]:
                outputfile.write('##SNPs in '+'\t'.join(info) + '\n')

            outputfile.write(
                '##Syn=Synonymous NS=Non-Synonymous'+'\n')

            header = ['##Coordinates', 'REF', 'SNP', 'Depth', 'Nb of reads REF', 'Nb reads SNPs', 'Ratio', 'Quality', 'Annotation', 'Product',
                        'Orientation', 'Coordinates in gene', 'Ref codon', 'SNP codon', 'Ref AA', 'SNP AA', 'Coordinates protein', 'Effect', 'Location']

            outputfile.write('\t'.join(header) + '\n')

            for eachData in all_data_collection:
                data_collection_sorted = sorted(
                    eachData, key=itemgetter(0))
                for eachSNP in data_collection_sorted:
                    outputfile.write(
                        '\t'.join([format(eachElem) for eachElem in eachSNP]) + '\n')

            outputfile.close()


        else:
            warnings.append(
                '{}: No SNPs were detected based on the specified filtering criteria'.format(File_name))

    if len(warnings) > 0:
        for warning in warnings:
            logger.warning(warning)

def combine(options):
    regions_to_exclude = []
    if options.exclude is not None:
        import yaml
        if Path(options.exclude).exists():
            with open(options.exclude, 'r') as fh:
                try:
                    regions_to_exclude = yaml.load(fh, Loader=yaml.FullLoader)
                except yaml.YAMLError as exc:
                    logger.error(exc)
        else:
            logger.error('The file ' + options.exclude + ' does not exist')
            sys.exit(0)

    prefix = "_snpToolkit_SNPs.txt"
    choices = {'ns': 'Non-synonymous', 's': 'Sysnonymous',
               'inter': 'intergenic', 'all': 'SNPs'}

    checkOutput = Path(choices[options.snps] + '_polymorphic_sites.txt')
    if checkOutput.exists():
        logger.warning(choices[options.snps] + '_polymorphic_sites.txt exists already and was created on ' + time.ctime(
            checkOutput.stat().st_ctime) + '. This file will be replaced. Press any key to continue or ctrl-c to exit!')
        try:
            input()
        except KeyboardInterrupt:
            sys.exit(0)

    if options.bamFilter != None:
        BamFilesToInclude = checksnpToolkitBam(options.bamFilter[2])

        if len(BamFilesToInclude) == 0:
            logger.error('None of the bam files names in the folder {} correspond to any of the snpToolkit outputs'.format(
                options.bamFilter[2]))
            sys.exit(0)
        else:
            BamBai = checkBamindex(options.bamFilter[2])
            if len(BamBai) > 0:
                logger.error('index file missing for: \n' + '\n'.join(BamBai))
                sys.exit(0)

    All_snpToolkitFiles = glob.glob("*" + prefix)
    snpToolkitFiles = []
    for eachFile in All_snpToolkitFiles:
        if os.stat(eachFile).st_size == 0:
            logger.warning(eachFile + ' is empty and will not be processed!')
        else:
            snpToolkitFiles.append(eachFile)

    if len(snpToolkitFiles) == 0:
        logger.error('No snpToolkit output files were detected')
        sys.exit(0)

    elif len(snpToolkitFiles) == 1:
        logger.error('At least two snpToolkit output files must be detected')
        sys.exit(0)

    else:
        logger.info('Searching for polymorphic sites...')
        SampleNames = []
        polymorphic_sites = []
        for eachFile in snpToolkitFiles:
            SampleNames.append(ntpath.basename(eachFile).split(prefix)[0])
            with open(eachFile, 'r') as f:
                for l in f.readlines():
                    if l[:2] != '##':
                        content = l.strip().split('\t')
  
                        if float(content[6]) >= float(options.ratio):
                            snp = [int(content[0])] + \
                                content[1:3] + content[8:-1]
                            if snp not in polymorphic_sites:
                                if options.location in l.split('\t')[-1] or options.location == l.split('\t')[-1]:
                                    if options.snps == 'all':
                                        polymorphic_sites.append(snp)
                                    elif options.snps == 's':
                                        if l.split('\t')[17] == 'Syn':
                                            polymorphic_sites.append(snp)
                                    elif options.snps == 'ns':
                                        if l.split('\t')[17] == 'NS':
                                            polymorphic_sites.append(snp)
                                    else:
                                        if l.split('\t')[8] == 'intergenic':
                                            polymorphic_sites.append(snp)

        if len(polymorphic_sites) == 0:
            logger.error(
                'No polymorphic sites found! are you sure you provided the right name when using the --location option?')
            sys.exit(0)

        polymorphic_sites.sort()
        header = '##ID Coordinates REF SNP Location Product Orientation NucPosition REF-codon NEW-codon REF-AA NEW-AA ProPostion Type'.split(
            ' ')
        infos = ['##snpToolkit=version '+__version__, '##commandline= ' + ' '.join(
            sys.argv[:]), '##location='+options.location, '##Number of polymorphic sites= ' + str(len(polymorphic_sites)), '\t'.join(header + SampleNames)]

        logger.info(choices[options.snps] +
                    ' polymorphic sites distribution. Please wait...')

        if options.bamFilter == None:
            distribution_result = snp_distribution(
                polymorphic_sites, snpToolkitFiles, regions_to_exclude)

        else:
            
            distribution_result = snp_distribution_missing(
                options.location, options.ratio, options.snps, polymorphic_sites, snpToolkitFiles, options.bamFilter[2], BamFilesToInclude, int(options.bamFilter[0]),  float(options.bamFilter[1]),regions_to_exclude)

        logger.info(
            'Creating ' + choices[options.snps] + '_polymorphic_sites.txt')
        outputfile1 = open(choices[options.snps] +
                           '_polymorphic_sites.txt', 'w')

        for info in infos:
            outputfile1.write(info + '\n')

        reference = ''
        i = 1
        for eachSNP in distribution_result:
            reference = reference + eachSNP[1]
            outputfile1.write('snp{}'.format(i) + '\t' +
                              '\t'.join([str(x) for x in eachSNP]) + '\n')
            i += 1
        outputfile1.close()

        logger.info('Creating ' + choices[options.snps] + '_alignment.fasta')
        outputfile2 = open(choices[options.snps] + '_alignment.fasta', 'w')
        outputfile2.write('>' + options.location + '\n' + reference + '\n')
        
        reconstruct_fasta(outputfile2, 13, SampleNames, distribution_result)


def expand(options):
    regions_to_exclude = []
    position_to_exclude=[]
    locations_to_exclude=[]
    if options.exclude is not None:
        import yaml
        if Path(options.exclude).exists():
            with open(options.exclude, 'r') as fh:
                try:
                    regions_to_exclude = yaml.load(fh, Loader=yaml.FullLoader)
                    position_to_exclude = regions_to_exclude['COORDINATES'].split(';')
                    locations_to_exclude = regions_to_exclude['KEYWORDS'].split(';')
                except yaml.YAMLError as exc:
                    logger.error(exc)
                
        else:
            logger.error('The file ' + options.exclude + ' does not exist')
            sys.exit(0)


    FilesToAdd = [FILE for FILE in glob.glob(
        options.directory + '/*_snpToolkit_SNPs.txt')]
    NewaDNA = [FILE for FILE in glob.glob(
        options.directory + '/*.bam')]
    with open(options.polymorphic_sites, 'r') as input:
        if position_to_exclude !=[]:
            PolyMorphicSites = [l.strip().split('\t') for l in input if '##' not in l and l.strip().split('\t')[1] not in position_to_exclude and l.strip().split('\t')[4] not in  position_to_exclude]    
        else:
            PolyMorphicSites = [l.strip().split('\t') for l in input if '##' not in l]  

    with open(options.polymorphic_sites, 'r') as input:
        info = [l.strip().split('\t') for l in input if '##ID' in l]


    header =info[0][14:] 

    if options.bamfiles_directory != None:
        aDNA = [FILE.split('.')[0].split('/')[-1] for FILE in glob.glob(options.bamfiles_directory + '/*.bam')]

    for i in tqdm(range(len(FilesToAdd)), ascii=True, desc='progress'):
        with open(FilesToAdd[i], 'r') as input:
            listSNPs=[l.strip().split('\t') for l in input if '##' not in l and options.location in l and l.strip().split('\t') [0] not in position_to_exclude and l.strip().split('\t') [8] not in locations_to_exclude ]

        for snp in PolyMorphicSites:
            if  snp[1] in [s[0] for s in listSNPs ]:
                snp.append ('1')
            else:
                bamFile = pysam.AlignmentFile(FilesToAdd[i].split('_snpToolkit')[0]+'.bam')
    
                try:
                    depthATposition = 0
                    # bamFile.count_coverage() return four array.arrays of the same length in order A C G T
                    NucleotidesOrder = 'ACGT'
                    NumberOfReads = []
                    k = 0
                    for eachCov in bamFile.count_coverage(options.location, int(snp[1])-1, int(snp[1])):
                        depthATposition = depthATposition + \
                            list(eachCov)[0]
                        NumberOfReads.append(
                            (list(eachCov)[0], NucleotidesOrder[k]))
                        k += 1
                    NumberOfReads.sort(reverse=True)
                    if depthATposition < int(options.cutoff):
                        snp.append('?')
                    else:
                        snp.append('0')
                except ValueError as e:
                    logging.error(
                        'Please use samtools view -H on this bam file to check the exact name.')
        done =[]
        bamfiles_list = list(glob.glob(options.bamfiles_directory+'*.bam'))
        for snp in listSNPs:          
            if snp[0] not in done:
                
                if snp[0] not in  [s[1] for s in PolyMorphicSites]:
                    distribution=[]
                    
                    
                    for x in range (len(header)):
                        flag = False
                        for file in bamfiles_list:
                            if header[x] in file:
                                bamFile = pysam.AlignmentFile(file)
                                flag = True
                                
                                try:
                                    depthATposition = 0
                                    # bamFile.count_coverage() return four array.arrays of the same length in order A C G T
                                    NucleotidesOrder = 'ACGT'
                                    NumberOfReads = []
                                    k = 0
                                    for eachCov in bamFile.count_coverage(options.location, int(snp[0])-1, int(snp[0])):
    
                                        depthATposition = depthATposition + list(eachCov)[0]
                                        NumberOfReads.append((list(eachCov)[0], NucleotidesOrder[k]))
                                        k += 1
                                    NumberOfReads.sort(reverse=True)
                                    
                                    if depthATposition < int(options.cutoff):
                                        distribution.append('?')
                                    else:
                                        if NumberOfReads[0][1] == snp[1]:
                                            distribution.append('0')
                                        else:
                                            distribution.append('1')
                                except ValueError as e:
                                    logging.error(
                                        'Please use samtools view -H on this bam file to check the exact name.')
                        if flag == False:
                            distribution.append ('0')
                    distribution.append ('1')
                    PolyMorphicSites.append (['SNP']+snp[:3]+snp[8:18]+distribution)
            done.append (snp[0])
                                    
        
        header.append(FilesToAdd[i].split('_snpToolkit_')[0].split('/')[-1])
    #TODO:add header for polymorphic sites file
    header_info="##ID,Coordinates,REF,SNP,Location,Product,Orientation,NucPosition,REF-codon,NEW-codon,REF-AA,NEW-AA,ProPostion,Type"
    with open (options.output+'-polymorphic_sites.txt', 'w') as snpoutput:
        snpoutput.write('\t'.join(header_info.split(',')+header)+'\n')
        for snp in PolyMorphicSites:
            snpoutput.write('\t'.join(snp)+'\n')
        



    with open(options.output+'-reconstracted.fasta', 'w') as fastaoutput:
        reference=''
        for SNP in PolyMorphicSites:
            reference = reference + SNP[2]
        
        fastaoutput.write('>'+options.location+'\n'+reference+'\n')
        i = 0
        while i < len(header):
            
            sequence = ''
            for SNP in PolyMorphicSites:

                if SNP[i+14] == '0':
                    sequence = sequence + SNP[2]
                elif SNP[i+14] == '1':
                    sequence = sequence + SNP[3]
                else:
                    sequence = sequence + '?'

            fastaoutput.write('>'+header[i]+'\n'+sequence+'\n')
            i +=1




#!/usr/bin/env python

# Copyright Amine Namouchi
# snpToolkit is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but without
# any warranty; without even the implied warranty of merchantability or fitness
# for a particular purpose. See the GNU General Public License for more details
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

__licence__ = 'GPLv3'
__author__ = 'Amine Namouchi'
__author_email__ = 'amine.namouchi@gmail.com'
__version__ = '2.2.3'

import argparse
import logging
import coloredlogs


def get_options():
    copyright = __licence__ + " licence | " + __author__ + " | " + __author_email__

    description = """
    snpToolkit can takes vcf files, as well as bam files (optional) as inputs. The vcf files could be generated using samtools/bcftools, gatk HaplotypeCaller or freeBayes.
    Please visit https://snptoolkit.readthedocs.io/en/latest/index.html for more information.
    """
    parser = argparse.ArgumentParser(
        description=description, formatter_class=argparse.RawTextHelpFormatter, epilog=copyright)

    subparsers = parser.add_subparsers(help='commands', dest='command')


    explore = subparsers.add_parser(
        name='explore', help='Explore your vcf files before annotation', epilog=copyright)

    requiredOptions1 = explore.add_argument_group(
        'snpToolkit explore required options')

    requiredOptions1.add_argument('-i', required=True, type=str, dest='identifier',
                                  help='Provide the input vcf files')

    annotate = subparsers.add_parser(
        name='annotate', help='Annotate one or multiple vcf files', epilog=copyright)

    requiredOptions2 = annotate.add_argument_group(
        'snpToolkit annotate required options')

    requiredOptions2.add_argument('-i', required=True, type=str, dest='identifier',
                                  help='Provide a specific identifier to recognize the file(s) to be analyzed')

    requiredOptions2.add_argument('-g', required=True, type=str, dest='genbank',
                                  help='Provide a genbank file')

    AdditonalOptions2 = annotate.add_argument_group(
        'snpToolkit annotate additional options')
    
    AdditonalOptions2.add_argument('-p', required=False,  default=1, dest='processors', type=int,
                                   help='Number of vcf files to be annotated in parallel default value [1]')


    AdditonalOptions2.add_argument('-f', required=False, dest='excludeCloseSNPs', type=int,
                                   help='Exclude SNPs if the distance between them is lower then the specified window size in bp')

    AdditonalOptions2.add_argument('-q', required=False, default=20, type=int, dest='quality',
                                   help='Quality score to consider as a cutoff for variant calling. default value [20]')

    AdditonalOptions2.add_argument('-d', required=False, default=3, type=int,
                                   dest='depth',  help='Minimum depth caverage. default value [3]')

    AdditonalOptions2.add_argument('-r', required=False,
                                   dest='ratio', default=0.001, type=float, help='Minimum ratio that correspond to the number of reads that has the mutated allele / total depth in that particular position. default value [0]')

    AdditonalOptions2.add_argument('-e', required=False, type=str, dest='exclude',
                                   help='Please provide a tab file with genomic regions to exclude in this format: region	start stop. region must correspond to the same name(s) of chromsome or plasmids as indicated in the genbank file')

    combine = subparsers.add_parser(
        name='combine', help='Identify polymorphic sites and create distribution table and alignment file', epilog=copyright)

    requiredOptions3 = combine.add_argument_group(
        'snpToolkit combine required options')

    requiredOptions3.add_argument('--loc', required=True, type=str, dest='location',
                                  help='Please provide for example the name of the chromosome or plasmid you want to create fasta alignemnt for')

    AdditonalOptions3 = combine.add_argument_group(
        'snpToolkit additional options')

    AdditonalOptions3.add_argument(
        '-r', required=False, type=float, dest='ratio', default=0.01, help='new versus reference allele ratio to filter SNPs from snpToolkit outputs. default [0]')

    AdditonalOptions3.add_argument('--bam', required=False, type=str,
                                   dest='bamFilter', nargs=3, help='provide the depth, ratio and the path to the folder containing the bam files. eg. 3 0.9 path')

    AdditonalOptions3.add_argument('--snps', required=False, type=str, dest='snps', default='all', choices=['ns', 's', 'all', 'inter'],
                                   help='Specify if you want to concatenate all SNPs or just synonymous (s), non-synonymous (ns) or intergenic (inter) SNPs. default [all]')

    AdditonalOptions3.add_argument('-e', required=False, type=str, dest='exclude',
                                   help='provide a yaml file with keywords and coordinates to be excluded') 



    viz = subparsers.add_parser(
        name='viz', help='visualize snptoolkit output files', epilog=copyright)

    requiredOptions4 = viz.add_argument_group('snpToolkit viz required options')
    requiredOptions4.add_argument('--dir', required=True, type=str, dest='directory',
                                  help='provide the path of the directory containing snptoolkit SNPs output files')


    expand = subparsers.add_parser(
        name='expand', help='expand existent list of polymorphic sites when new SNP output files are availble', epilog=copyright)

    requiredOptions4 = expand.add_argument_group('snpToolkit viz required options')
    requiredOptions4.add_argument('--dir', required=True, type=str, dest='directory',
                                  help='provide the path to the directory containing snptoolkit output files to be added')
    requiredOptions4.add_argument('-p', required=True, type=str, dest='polymorphic_sites',
                                help='provide the polymorphic sites file already generated')
    requiredOptions4.add_argument('-o', required=True, type=str, dest='output',
                                help='output')
    requiredOptions4.add_argument('-l', required=True, type=str, dest='location',
                                help='name of chromosome, plasmid, contig')
    requiredOptions4.add_argument('--bam', required=False, type=str, dest='bamfiles_directory',
                                help='bam files directory of already analysed anciant DNA')
    requiredOptions4.add_argument('-c', required=False, type=str, dest='cutoff',
                                help='cutoff  of coverage if bam file is provided ')
    requiredOptions4.add_argument('-e', required=False, type=str, dest='exclude',
                                help='exclude yaml ')

    return parser.parse_args()

def setupLogger():

    logger = logging.getLogger(__name__)
    coloredlogs.install(logger=logger, fmt='[%(asctime)s] [%(levelname)s] [%(message)s]', datefmt='%H:%M:%S', field_styles={
                        'levelname': {'color': 'white', 'bold': False, 'background': 'black'},  'asctime': {'color': 'green'}})
    filehandle = logging.FileHandler('snpToolkit.log')
    formatter = logging.Formatter(
        fmt='[%(asctime)s] [%(levelname)s] [%(message)s]', datefmt='%H:%M:%S')
    filehandle.setFormatter(formatter)
    logger.addHandler(filehandle)
    return logger



#!/usr/bin/env python

# Copyright Amine Namouchi
# snpToolkit is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but without
# any warranty; without even the implied warranty of merchantability or fitness
# for a particular purpose. See the GNU General Public License for more details
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.


__licence__ = 'GPLv3'
__author__ = 'Amine Namouchi'
__author_email__ = 'amine.namouchi@gmail.com'
__version__ = '2.2.3'


import glob
import logging
import coloredlogs
import argparse
import plotly.graph_objs as go
import plotly.offline as pyo
import plotly.figure_factory as ff
from plotly.colors import n_colors
import dash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies  import Input, Output
import dash_table as dt
from itertools import islice
import pandas as pd 




#TODO: adjust the size of the input text. color dots in scatter according to SNPS effect.


parser = argparse.ArgumentParser()
parser.add_argument('--dir', dest='directory', required=True, help='provide the path to the directory containing snptoolkit output files')
args= parser.parse_args()

list_files = glob.glob(args.directory+'/*_snpToolkit_SNPs.txt')
snptoolkitInputs =[]
for f in list_files:
    snptoolkitInputs.append ({'label':f.split('/')[-1].split('_snpToolkit_')[0],'value':f.split('/')[-1].split('_snpToolkit_')[0]})


colors={'background':'#262626','text':'#ffffff','textHeader':'#111111'}
app = dash.Dash()


header = ['##Coordinates', 'REF', 'SNP', 'Ref codon', 'SNP codon', 'Ref AA', 'SNP AA','Effect','Annotation', 'Product',
            'Orientation', 'Coordinates in gene',  'Coordinates protein','Depth', 'Nb of reads REF', 'Nb reads SNPs', 'Ratio', 'Quality']

app.layout = html.Div(
    [
        html.Div([
        dcc.Dropdown(
            id='name-dropdown',
            options=snptoolkitInputs,
            value = snptoolkitInputs[0]['value']
            ),
            ],style={'width': '50%', 'display': 'inline-block'}),
        html.Div([
        dcc.Dropdown(
            id='opt-dropdown',
            ),
            ],style={'width': '50%', 'display': 'inline-block'}
        ),
        html.Hr(),
        dcc.RadioItems(id='selected-type',options=[
            {'label': 'All SNPs', 'value': 'ALL'},
            {'label': 'Synonymous SNPs', 'value': 'SYN'},
            {'label': 'Non-synonymous SNPs', 'value': 'NS'},
            ],value='ALL',labelStyle={'display': 'inline-block'}
        ),
        #TODO: dcc.Input(placeholder='Enter a window size to compute SNPs density...',type='number',id="windowsize",value=''),
        dcc.Graph(id='graph'),
        html.Hr(),
        html.Div(dt.DataTable(columns = [{"name": i, "id": i} for i in header],
        id='snpTable', 
        filter_action="native",
        page_size=20,
        style_table={ 'align':'center','overflowY': 'scroll'},
        style_cell={'textAlign': 'center', 'border': '1px solid grey'},
        style_header={
        'color':'white',
        'backgroundColor': 'rgb(62, 155, 184)',
        'border': '1px solid black'
        },
        style_data_conditional=[
            {
                'if': {
                'column_id': 'Effect',
                'filter_query': '{Effect} eq "NS"'
            },
            'backgroundColor': '#FBBF4C',
            'color': 'white',
        },
        {
                'if': {
                'column_id': 'Effect',
                'filter_query': '{Effect} eq "Syn"'
            },
            'backgroundColor': '#51A8C7',
            'color': 'white',
        },
        {
                'if': {
                'column_id': 'Effect',
                'filter_query': '{Effect} eq "."'
            },
            'backgroundColor': '#0e2f2e',
            'color': '#0e2f2e',
        }
        ])),
    ]
)
@app.callback(
    Output('opt-dropdown', 'options'),
    [Input('name-dropdown', 'value')]
)
def update_dropdown(name):
    SNPsLocation =[]
    for eachFile in list_files:
        if name in eachFile:
            with open (eachFile, 'r') as f:
                headerFile = islice(f, 30)
                for c in headerFile:
                    if '##SNPs in ' in c:  
                        SNPsLocation.append ({'label': c.strip().replace('##SNPs in ','').split('\t')[0], 'value': c.strip().replace('##SNPs in ','').split('\t')[0]})
    
    return SNPsLocation


@app.callback(
    Output('opt-dropdown', 'label'),
    [Input('opt-dropdown', 'options')]
)
def default_selection(option2):
    if len (option2) > 1:
        return ''
    else:
        return option2[0]['value']


@app.callback (Output('graph','figure'),[Input('name-dropdown','value'),Input('opt-dropdown','value'),Input('selected-type','value')])##TODO: Input('windowsize','value')
def update_graph(sample,location,snpType):
    for eachFile in list_files:
        if sample in eachFile:
            fh = open (eachFile,'r')
            line = fh.readline()
            x = 0
            while x !='':
                if line.startswith('##Coordinates')==True:
                    break
                x+=1
                line = fh.readline()
            df = pd.read_csv(eachFile,sep='\t',skiprows=x)

    df1 = df.loc[(df['Effect'] == 'NS') & (df['Location'] == location)][header]
    df2 = df.loc[(df['Effect'] == 'Syn') & (df['Location'] == location)][header]

    layout = go.Layout(title='Depth vs Ratio',xaxis={'title':'Coordinates'},yaxis={'title':'Depth'},hovermode='closest')
    if snpType == 'ALL':
        data = [go.Scatter(name="Others",x=df[(df["Location"]==location)]["##Coordinates"],y=df[(df["Location"]==location)]["Depth"],mode="markers",opacity=0.5,marker={"color":"grey"}),go.Scatter(name='NS',x=df1["##Coordinates"],y=df1["Depth"],mode="markers",opacity=0.8,marker={"color":"#FBBF4C"}),go.Scatter(name='Syn',x=df2["##Coordinates"],y=df2["Depth"],mode="markers",opacity=0.8,marker={"color":"#51A8C7"})]
    elif snpType == 'NS':
        
        data = [go.Scatter(x=df1["##Coordinates"],y=df1["Depth"],mode="markers",opacity=0.8,marker={"color":"#FBBF4C"})]
    else: 
        data = [go.Scatter(x=df2["##Coordinates"],y=df2["Depth"],mode="markers",opacity=0.8,marker={"color":"#51A8C7"})]
    return {'data':data,'layout':layout}


@app.callback (Output('snpTable','data'),[Input('name-dropdown','value'),Input('opt-dropdown','value')])
def update_table(sample,location):
    for eachFile in list_files:
        if sample in eachFile:
            fh = open (eachFile,'r')
            line = fh.readline()
            x = 0
            while x !='':
                if line.startswith('##Coordinates')==True:
                    break
                x+=1
                line = fh.readline()
            df = pd.read_csv(eachFile,sep='\t',skiprows=x)
    return df[(df['Location'] == location)][header].to_dict('rows')
        

if __name__ == '__main__':
    app.run_server()



#!/usr/bin/env python

# Copyright Amine Namouchi
# snpToolkit is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but without
# any warranty; without even the implied warranty of merchantability or fitness
# for a particular purpose. See the GNU General Public License for more details
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.


__licence__ = 'GPLv3'
__author__ = 'Amine Namouchi'
__author_email__ = 'amine.namouchi@gmail.com'
__version__ = '2.2.3'


import gzip
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from itertools import islice
from pathlib import Path
import glob
import ntpath
import sys
import logging
import coloredlogs
from tqdm import tqdm


standcode = {
    'ttt': 'F', 'tct': 'S', 'tat': 'Y', 'tgt': 'C', 'ttc': 'F', 'tcc': 'S',
    'tac': 'Y', 'tgc': 'C', 'tta': 'L', 'tca': 'S', 'taa': '#', 'tga': '#',
    'ttg': 'L', 'tcg': 'S', 'tag': '#', 'tgg': 'W', 'ctt': 'L', 'cct': 'P',
    'cat': 'H', 'cgt': 'R', 'ctc': 'L', 'ccc': 'P', 'cac': 'H', 'cgc': 'R',
    'cta': 'L', 'cca': 'P', 'caa': 'Q', 'cga': 'R', 'ctg': 'L', 'ccg': 'P',
    'cag': 'Q', 'cgg': 'R', 'att': 'I', 'act': 'T', 'aat': 'N', 'agt': 'S',
    'atc': 'I', 'acc': 'T', 'aac': 'N', 'agc': 'S', 'ata': 'I', 'aca': 'T',
    'aaa': 'K', 'aga': 'R', 'atg': 'M', 'acg': 'T', 'aag': 'K', 'agg': 'R',
    'gtt': 'V', 'gct': 'A', 'gat': 'D', 'ggt': 'G', 'gtc': 'V', 'gcc': 'A',
    'gac': 'D', 'ggc': 'G', 'gta': 'V', 'gca': 'A', 'gaa': 'E', 'gga': 'G',
    'gtg': 'V', 'gcg': 'A', 'gag': 'E', 'ggg': 'G'
}

Reverse_nuc = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}


def parse_genbank_file(gb_filename):
    """
    parse_genbank_file takes as input file a genbank file and retreive for each entry (chromosome, plasmids)
    all data relative to each gene, rRNA, tRNA, ncRNA as follow:
    gene|synonymous  type  start   stop  orientation   product  nucleotide_sequence   protein_sequence
    """

    gbank = SeqIO.parse(open(gb_filename, "r"), "genbank")

    locus_order = []
    collectionDB = {}
    for genome in gbank:
        locus_order.append(genome.id)
        whole_nuc_seq=genome.seq
        collect_data = []
        for feature in genome.features:
            if(feature.type == "CDS"):
                if feature.location_operator == 'join':
                    #TODO: here it is important to pass the coordinates of the fragsment to join and the the other intromic' part. this is important as it is necessray to change the coordiantes of the SNPs according to the new generated sequence. 
                    #need to work more with the example sample1_n50_gatk.vcf
                    nucseq=''
                    for sub_location in feature.location.parts:
                        if sub_location.strand == 1:
                            nucseq = nucseq + whole_nuc_seq[sub_location.start:sub_location.end]
                else:
                    if 'gene' in feature.qualifiers.keys() and 'pseudo' not in feature.qualifiers.keys():
                        try:
                            locus = feature.qualifiers['locus_tag'][0]
                        except KeyError:
                            try:
                                locus = feature.qualifiers['gene_synonym'][0]
                            except KeyError:
                                locus = feature.qualifiers['gene'][0]
                        gene = feature.qualifiers['gene'][0]
        
                        start = format(feature.location.start.position + 1)
                        stop = format(feature.location.end.position)
                        strand = feature.location.strand
                        product = feature.qualifiers['product'][0]
                        nucseq = feature.extract(genome.seq)
                        if strand == 1:
                            collect_data.append(
                                [locus + '|' + gene, 'CDS', start, stop, '+', product, format(nucseq)])
                        else:
                            collect_data.append(
                                [locus + '|' + gene, 'CDS', start, stop, '-', product, format(nucseq)])
                    else:
                        if 'pseudo' not in feature.qualifiers.keys():
                            try:
                                locus = feature.qualifiers['locus_tag'][0]
                            except KeyError:
                                try:
                                    locus = feature.qualifiers['gene_synonym'][0]
                                except KeyError:
                                    locus = feature.qualifiers['gene'][0]
                            gene = feature.qualifiers['locus_tag'][0]
                            start = format(feature.location.start.position + 1)
                            stop = format(feature.location.end.position)
                            strand = feature.location.strand
                            product = feature.qualifiers['product'][0]
                            nucseq = feature.extract(genome.seq)
                            if strand == 1:
                                collect_data.append(
                                    [locus + '|' + gene, 'CDS', start, stop, '+', product, format(nucseq)])
                            else:
                                collect_data.append(
                                    [locus + '|' + gene, 'CDS', start, stop, '-', product, format(nucseq)])
            else:

                if feature.type in ['tRNA', 'rRNA', 'ncRNA']:
                    try:
                        ID1 = feature.qualifiers['locus_tag'][0]
                        ID2 = feature.qualifiers['locus_tag'][0]
                    except KeyError:
                        try:
                            ID1 = feature.qualifiers['gene'][0]
                            ID2 = feature.qualifiers['gene'][0]
                        except KeyError:
                            ID1=feature.type
                            ID2=feature.type
                    type = feature.type
                    start = format(feature.location.start.position + 1)
                    stop = format(feature.location.end.position)
                    strand = feature.location.strand
                    if 'product' in feature.qualifiers.keys():
                        product = feature.qualifiers['product'][0]
                    else:
                        if 'note' in feature.qualifiers.keys():
                            product = feature.qualifiers['note'][0]
                    nucseq = feature.extract(genome.seq)
                    if strand == 1:
                        collect_data.append(
                            [ID1 + '|' + ID2, type, start, stop, '+', product, format(nucseq), '*'])
                    else:
                        collect_data.append(
                            [ID1 + '|' + ID2, type, start, stop, '-', product, format(nucseq), '*'])

        collectionDB[genome.id + ': ' + genome.description +
                     ' {} bp'.format(len(genome.seq))] = collect_data
    collectionDB['locus_order'] = locus_order
    if collectionDB == {}:
        logging.error('Please check your genbank file')
        sys.exit(0)
    return collectionDB


def isVCF(filename):

    if filename.endswith(".vcf.gz"):
        filehandle = gzip.open(filename, 'rb')
        line = filehandle.readline()
        if b'##fileformat=VCF' in line:
            return True
        else:
            return False

    elif filename.endswith(".vcf"):
        filehandle = open(filename, 'r')
        line = filehandle.readline()
        if '##fileformat=VCF' in line:
            return True
        else:
            return False


class VCFtoolBox(object):
    """
    This class will handle vcf files to:
        1. validates if a given file is a valid vcf file
        2. distinguish how the vcf file was generated using mpileup or gatk
        3. extract all single nucleotides polymorphism
        4. extract all insertion and deletions
    """

    def __init__(self, filename):
        self.file_name = filename
        if self.file_name.endswith('.vcf'):
            self.fh = open(self.file_name, 'r')
            self.file_type = 'string'
        else:
            if self.file_name.endswith('.vcf.gz'):
                self.fh = gzip.open(self.file_name, 'rb')
                self.file_type = 'bytes'
            else:
                self.fh = zlib.decompress.open(self.file_name, 'rb')
                self.file_type = 'bytes'

    def vcf_generator(self):
        if self.file_type == 'string':
            for line in islice(self.fh, 10):
                if 'mpileup' in line:
                    return 'samtools'
                elif 'GATK' in line:
                    return 'GATK'
                elif 'freeBayes' in line:
                    return 'freeBayes'
                else:
                    pass
            return '%s will be skipped as it seems to be not a valid vcf file or generated with a program different than samtools, gatk or freeBayes!' % self.file_name

        else:
            for line in islice(self.fh, 10):
                if b'mpileup' in line:
                    return 'samtools'
                elif b'GATK' in line:

                    return 'GATK'
                elif b'freeBayes' in line:
                    return 'freeBayes'
                else:
                    pass
            return '%s will be skipped as it seems to be not a valid vcf file or generated with a program different than samtools, gatk or freeBayes!' % self.file_name

    def extract_all_snps(self):
        dna = ('A', 'C', 'T', 'G')
        if self.file_type == 'string':
            list_snps = [line.strip().split('\t') for line in self.fh if '##' not in line and 'REF	ALT' not in line and line.split(
                '\t')[3] in dna and line.split('\t')[4] in dna and line.split('\t')[3] != line.split('\t')[4]]
        else:
            list_snps = [str(line, 'utf-8').strip().split('\t') for line in self.fh if '##' not in str(line, 'utf-8') and 'REF	ALT' not in str(line, 'utf-8') and str(
                line, 'utf-8').split('\t')[3] in dna and str(line, 'utf-8').split('\t')[4] in dna and str(line, 'utf-8').split('\t')[3] != str(line, 'utf-8').split('\t')[4]]

        return list_snps

    def extract_all_indels(self):
        if self.file_type == 'string':
            list_indels = [line.strip().split('\t') for line in self.fh if '##' not in line and 'REF	ALT' not in line and 'INDELS' in line and line.split(
                '\t')[3] != line.split('\t')[4] and line.split('\t')[6] == 'PASS']
        else:
            list_indels = [str(line, 'utf-8').strip().split('\t') for line in self.fh if '##' not in str(line, 'utf-8') and 'REF	ALT' not in str(line, 'utf-8') and 'INDEL' in str(line, 'utf-8') and
                           str(line, 'utf-8').split('\t')[3] != str(line, 'utf-8').split('\t')[4] and str(line, 'utf-8').split('\t')[6] == 'PASS']

        return list_indels

    def extract_all_variation(self):
        if self.file_type == 'string':
            list_indels = [line.strip().split('\t') for line in self.fh if '##' not in line and 'REF	ALT' not in line and line.split(
                '\t')[3] != line.split('\t')[4]]
        else:
            list_indels = [str(line, 'utf-8').strip().split('\t') for line in self.fh if '##' not in str(line, 'utf-8') and 'REF	ALT' not in str(line, 'utf-8') and
                           str(line, 'utf-8').split('\t')[3] != str(line, 'utf-8').split('\t')[4]]

        return list_indels


class SNPfiltering (object):
    """
    This class allow:
        1. check the presence of specific SNP among all extracted SNPs
        2. Exclude SNPs if the distance between them is lower than a defined distance
        3. Exclude SNPs if they are located in regions specified by the user
        4. Exclude SNPs if the distance between them is lower than a defined distance and if they are located in regions specified by the user
    """

    def __init__(self, snps):
        self.snps = snps

    def check_presence(self, location, coordinate):
        for eachElem in self.snps:
            if location in eachElem[0] and coordinate == int(eachElem[1]):
                reference = eachElem[3]
                variant = eachElem[4]
                quality = eachElem[5]
                return '%s: %s%s%s' % (location, reference, str(coordinate), variant)
        return 'No SNP found in %s at posSNPition %d' % (location, coordinate)

    def FilterCloseSNPs(self, distance):
        """
        SNP_coordinates is set by defaut to 1 by reference to vcf files. this method will be used in case of tab input file too. user just set the column where the snp coordinates are located.
        distance refer to the distance betwen to adjacent SNPs
        """
        filteredSNPs = []

        i = 0
        while i < len(self.snps) - 1:
            if i == 0:
                if int(self.snps[i + 1][1]) - int(self.snps[i][1]) > distance:
                    filteredSNPs.append(self.snps[i])
            elif i == len(self.snps) - 1:
                if int(self.snps[i + 1][1]) - int(self.snps[i][1]) > distance:
                    filteredSNPs.append(self.snps[i])
                    filteredSNPs.append(self.snps[i + 1])
            else:
                if int(self.snps[i][1]) - int(self.snps[i-1][1]) > distance and int(self.snps[i + 1][1]) - int(self.snps[i][1]) > distance:
                    filteredSNPs.append(self.snps[i])

            i += 1
        return filteredSNPs

    def FilterSNPtoExclude(self, coordinates_to_exclude):
        filteredSNPs = []
        for content in self.snps:
            flag = False
            for eachRegion in coordinates_to_exclude:
                if content[0] == eachRegion[0] and int(eachRegion[2]) >= int(content[1]) >= int(eachRegion[1]):
                    flag = True
                    break
            if flag == False:
                filteredSNPs.append(content)
        return filteredSNPs
        
    def FilterCloseSNPsandInREgions(self, distance, coordinates_to_exclude):
        filteredSNPs1 = []
        for content in self.snps:
            flag = False
            for eachRegion in coordinates_to_exclude:
                if content[0] == eachRegion[0] and int(eachRegion[2]) >= int(content[1]) >= int(eachRegion[1]):
                    flag = True
                    break
            if flag == False:
                filteredSNPs1.append(content)

        filteredSNPs2 = []
        i = 0
        while i < len(filteredSNPs1) - 1:
            if i == 0:
                if int(filteredSNPs1[i + 1][1]) - int(filteredSNPs1[i][1]) > distance:
                    filteredSNPs2.append(filteredSNPs1[i])
            elif i == len(filteredSNPs1) - 1:
                if int(filteredSNPs1[i + 1][1]) - int(filteredSNPs1[i][1]) > distance:
                    filteredSNPs2.append(filteredSNPs1[i])
                    filteredSNPs2.append(filteredSNPs1[i + 1])
            else:
                if int(filteredSNPs1[i + 1][1]) - int(filteredSNPs1[i][1]) > distance and int(filteredSNPs1[i][1]) - int(filteredSNPs1[i - 1][1]) > distance:
                    filteredSNPs2.append(filteredSNPs1[i])
            i += 1

        return filteredSNPs1, filteredSNPs2


class SNPselect (object):

    """
    This class allow to select SNPs based on their quality and depth. It can handle data from mpileup, gatk or freebayes
    """

    def __init__(self, snps, QUAL, DEPTH, RATIO):
        self.snps = snps
        self.QUAL = QUAL
        self.DEPTH = DEPTH
        self.RATIO = RATIO

    def ExtractSNPinfo(self, origin):
        count_QUAL = 0
        count_DEPTH = 0
        count_RATIO = 0
        data_collection = {}
        units = list(set([elem[0] for elem in self.snps]))
        for eachUnit in units:
            extracted_data = []
            for content in self.snps:
                if eachUnit == content[0]:
                    coordiantes = int(content[1])
                    REF = content[3]

                    ALT = content[4]
                    quality = float(content[5])
                    if quality >= self.QUAL:
                        count_QUAL += 1
                        varInfo = content[7].split(';')
                        if origin == 'samtools':
                            for eachElem in varInfo:
                                if 'DP=' in eachElem:
                                    depth = int(eachElem.split('=')[1])
                                elif 'DP4=' in eachElem:
                                    DP4 = eachElem.split('=')[1].split(',')
                                    nbREADref = int(DP4[0]) + int(DP4[1])
                                    nbREADalt = int(DP4[2]) + int(DP4[3])

                            if depth >= self.DEPTH:
                                count_DEPTH += 1
                                if nbREADref == 0:
                                    ratio = 1.0
                                    extracted_data.append([coordiantes, REF, ALT, depth, nbREADref, nbREADalt, float(
                                        ('%.2f') % ratio), quality])
                                    count_RATIO += 1
                                else:
                                    ratio = float(
                                        nbREADalt*1.0/(nbREADref+nbREADalt))
                                    if ratio >= self.RATIO:
                                        extracted_data.append([coordiantes, REF, ALT, depth, nbREADref, nbREADalt, float(
                                            ('%.2f') % ratio), quality])
                                        count_RATIO += 1
                        elif origin == 'GATK':
                            for eachElem in varInfo:
                                if 'DP=' in eachElem:
                                    depth = int(eachElem.split('=')[1])
                            snpInfo = content[9].split(':')[1].split(',')
                            nbREADref = int(snpInfo[0])
                            nbREADalt = int(snpInfo[1])
                           
                            if depth >= self.DEPTH:
 
                                count_DEPTH += 1
                                if nbREADref == 0:
                                    ratio = 1.0
                                    extracted_data.append([coordiantes, REF, ALT, depth, nbREADref, nbREADalt, float(
                                        ('%.2f') % ratio), quality])
                                    count_RATIO += 1
                                else:
                                    ratio = float(
                                        nbREADalt*1.0/(nbREADref+nbREADalt))
                                    if ratio >= self.RATIO:
                                        extracted_data.append([coordiantes, REF, ALT, depth, nbREADref, nbREADalt, float(
                                            ('%.2f') % ratio), quality])
                                        count_RATIO += 1

                        else:  # freeBayes
                            for eachElem in varInfo:
                                if 'DP=' in eachElem:
                                    depth = int(eachElem.split('=')[1])
                                elif 'SRF=' in eachElem:
                                    nbREADref_F = int(eachElem.split('=')[1])
                                elif 'SRR=' in eachElem:
                                    nbREADref_R = int(eachElem.split('=')[1])
                                elif 'SAF=' in eachElem:
                                    nbREADalt_F = int(eachElem.split('=')[1])
                                elif 'SAR=' in eachElem:
                                    nbREADalt_R = int(eachElem.split('=')[1])

                            nbREADref = nbREADref_F + nbREADref_R
                            nbREADalt = nbREADalt_F + nbREADalt_R

                            if depth >= self.DEPTH:
                                count_DEPTH += 1
                                if nbREADref == 0:
                                    ratio = 1.0
                                    extracted_data.append([coordiantes, REF, ALT, depth, nbREADref, nbREADalt, float(
                                        ('%.2f') % ratio), quality])
                                    count_RATIO += 1
                                else:
                                    ratio = float(
                                        nbREADalt*1.0/(nbREADref+nbREADalt))
                                    if ratio >= self.RATIO:
                                        extracted_data.append([coordiantes, REF, ALT, depth, nbREADref, nbREADalt, float(
                                            ('%.2f') % ratio), quality])
                                        count_RATIO += 1
            data_collection[eachUnit] = extracted_data
        return data_collection, count_QUAL, count_DEPTH, count_RATIO


class SNPannotation (object):
    """
    This class will handle SNPs lists Objects to map them according to a provided genbank file by the user
    """

    def __init__(self, filteredSNPs):
        self.filteredSNPs = filteredSNPs  # list of filtered SNPs

    def mapSNPs(self, GenbankContent):
        """
        this method will allow to map all provided SNPs according to their distribution in Gene,
        RBS, tRNA, rRNA, ncRNA and intergenic. In addition SNPs will be grouped according to their
        location in chromosome or plasmids. This information is retreived from the LOCUS tag in genbank file. The output of this method is a dictionary with the foloowing structure:
        {LOCUS:{'Genes':[snps......],'RBS':[snps.......], etc}}
        """
        data_collector = {}
        gbank_accessions = GenbankContent.keys()
        for eachKEY in self.filteredSNPs.keys():
            if len(self.filteredSNPs[eachKEY]) > 0:
                collected_snp = self.filteredSNPs[eachKEY]
                snps_in_genes = []
                snps_in_RBS = []
                snps_in_tRNA = []
                snps_in_rRNA = []
                snps_in_ncRNA = []
                snps_in_pseudogene = []
                intergenic = []
                for eachAccession in gbank_accessions:
                    if eachAccession.split(':')[0] in eachKEY:
                        annotationDB = GenbankContent[eachAccession]
                        for eachSNP in collected_snp:
                            flag = False
                            posSNP = eachSNP[0]
                            for eachElem in annotationDB:
                                if posSNP <= int(eachElem[3]) and posSNP >= int(eachElem[2]):
                                    flag = True
                                    if eachElem[1] == 'Gene':
                                        snps_in_genes.append(eachSNP)
                                    elif eachElem[1] == 'RBS':
                                        snps_in_RBS.append(eachSNP)
                                    elif eachElem[1] == 'tRNA':
                                        snps_in_tRNA.append(eachSNP)
                                    elif eachElem[1] == 'rRNA':
                                        snps_in_rRNA.append(eachSNP)
                                    elif eachElem[1] == 'ncRNA':
                                        snps_in_ncRNA.append(eachSNP)
                                    elif eachElem[1] == 'pseudogene':
                                        snps_in_pseudogene.append(eachSNP)
                            if flag == False:
                                intergenic.append(eachSNP)
                        data_collector[eachAccession] = {'Gene': snps_in_genes, 'tRNA': snps_in_tRNA, 'RBS': snps_in_RBS,
                                                         'rRNA': snps_in_rRNA, 'ncRNA': snps_in_ncRNA, 'pseudogene': snps_in_pseudogene, 'intergenic': intergenic}

        return data_collector

    def mapAndannoINDELS(self, GenbankContent):
        data_collector = {}
        gbank_accessions = GenbankContent.keys()
        for eachKEY in self.filteredSNPs.keys():
            if len(self.filteredSNPs[eachKEY]) > 0:
                mappedINDELS = []

                for eachAccession in gbank_accessions:
                    if eachAccession.split(':')[0] in eachKEY or eachKEY in eachAccession:
                        annotationDB = GenbankContent[eachAccession]
                        for eachINDEL in self.filteredSNPs[eachKEY]:
                            flag = False
                            posINDEL = eachINDEL[0]
                            reference_allele = eachINDEL[1]
                            mutated_allele = eachINDEL[2]
                            total_depth = eachINDEL[3]
                            nb_reads_reference = eachINDEL[4]
                            nb_reads_mutated = eachINDEL[5]
                            ratio = eachINDEL[6]
                            quality = eachINDEL[7]
                            if len(reference_allele) > len(mutated_allele):
                                event = 'deletion'
                                size_event = len(
                                    reference_allele) - len(mutated_allele)
                            else:
                                event = 'insertion'
                                size_event = len(
                                    mutated_allele)-len(reference_allele)
                            for eachElem in annotationDB:
                                if posINDEL <= int(eachElem[3]) and posINDEL >= int(eachElem[2]):
                                    flag = True

                                    type = eachElem[1]
                                    gene_name = eachElem[0]
                                    gene_function = eachElem[5]

                                    collected_data = eachINDEL + [type, gene_name, gene_function,
                                                                  event, size_event]
                                    # collected_data = [posINDEL, reference_allele, mutated_allele, event, size_event, total_depth,
                                    #     nb_reads_reference, nb_reads_mutated, ratio, quality, type, gene_name, gene_function]
                                    mappedINDELS.append(collected_data)

                            if flag == False:
                                mappedINDELS.append(
                                    eachINDEL + ['intergenic', '.', '.', event, size_event])
                                cutoff_distance = 100
                                if posINDEL < int(annotationDB[0][2]):
                                    if annotationDB[0][4] == '+':
                                        distance_gene = int(
                                            annotationDB[0][2]) - posINDEL
                                        if distance_gene <= cutoff_distance:
                                            mappedINDELS.append(eachINDEL + ['intergenic', '.', str(
                                                distance_gene) + ' bp from ' + annotationDB[0][0], event, size_event])
                                        else:
                                            mappedINDELS.append(
                                                eachINDEL + ['intergenic', '.', '.', event, size_event])

                                    else:
                                        mappedINDELS.append(
                                            eachINDEL +
                                            ['intergenic', '.', '.', event, size_event])
                                else:
                                    if posINDEL > int(annotationDB[-1][3]):
                                        if annotationDB[-1][4] == '-':
                                            distance_gene = posINDEL - \
                                                int(annotationDB[0][3])
                                            if distance_gene <= cutoff_distance:
                                                mappedINDELS.append(eachINDEL + ['intergenic', '.', '+', str(
                                                    distance_gene) + ' bp from ' + annotationDB[-1][0], '.'])
                                            else:
                                                mappedINDELS.append(
                                                    eachINDEL +
                                                    ['intergenic', '.',
                                                     '+', '.', '.'])

                                        else:
                                            mappedINDELS.append(
                                                eachINDEL +
                                                ['intergenic', '.',
                                                 '+', '.', '.'])

                                k = 0
                                while k < len(annotationDB) - 1:
                                    if int(annotationDB[k][3]) < posINDEL < int(annotationDB[k + 1][2]):
                                        if annotationDB[k][4] == '+' and annotationDB[k + 1][4] == '+':
                                            distance_gene = int(
                                                annotationDB[k + 1][2]) - posINDEL
                                            if distance_gene <= cutoff_distance:
                                                mappedINDELS.append(eachINDEL + ['intergenic', '.', '+', str(
                                                    distance_gene) + ' bp from ' + annotationDB[k + 1][0], '.'])
                                            else:
                                                mappedINDELS.append(
                                                    eachINDEL +
                                                    ['intergenic', '.',
                                                     '+', '.', '.'])

                                        elif annotationDB[k][4] == '-' and annotationDB[k + 1][4] == '+':
                                            distance_gene1 = posINDEL - \
                                                int(annotationDB[k][3])
                                            distance_gene2 = int(
                                                annotationDB[k + 1][2]) - posINDEL
                                            if distance_gene1 <= cutoff_distance and distance_gene2 <= cutoff_distance:
                                                mappedINDELS.append(eachINDEL + ['intergenic', '.', '+', str(distance_gene1) + ' bp from ' + annotationDB[k][0] + ' and ' + str(
                                                    distance_gene2) + ' bp from ' + annotationDB[k + 1][0], '.'])
                                            elif distance_gene1 <= cutoff_distance and distance_gene2 > cutoff_distance:
                                                mappedINDELS.append(eachINDEL + ['intergenic', '.', '+', str(
                                                    distance_gene1) + ' bp from ' + annotationDB[k][0], '.'])
                                            elif distance_gene2 <= cutoff_distance and distance_gene1 > cutoff_distance:
                                                mappedINDELS.append(eachINDEL + ['intergenic', '.', '+', str(
                                                    distance_gene2) + ' bp from ' + annotationDB[k + 1][0], '.'])
                                            else:
                                                mappedINDELS.append(
                                                    eachINDEL +
                                                    ['intergenic', '.',
                                                     '+', '.', '.'])

                                        elif annotationDB[k][4] == '-' and annotationDB[k + 1][4] == '-':
                                            distance_gene = posINDEL - \
                                                int(annotationDB[k][3])
                                            if distance_gene <= cutoff_distance:
                                                mappedINDELS.append(eachINDEL + ['intergenic', '.', '+', str(
                                                    distance_gene) + ' bp from ' + annotationDB[k][0], '.'])
                                            else:
                                                mappedINDELS.append(
                                                    eachINDEL +
                                                    ['intergenic', '.',
                                                     '+', '.', '.'])

                                        else:

                                            mappedINDELS.append(
                                                eachINDEL +
                                                ['intergenic', '.',
                                                 '+', '.', '.'])

                                    k += 1

                        data_collector[eachAccession] = mappedINDELS

        return data_collector

    def mapAndannoSNPs(self, GenbankContent):
        data_collector = {}
        gbank_accessions = GenbankContent.keys()
        for eachKEY in self.filteredSNPs.keys():
            if len(self.filteredSNPs[eachKEY]) > 0:
                snps_in_genes = []
                snps_in_RBS = []
                snps_in_tRNA = []
                snps_in_rRNA = []
                snps_in_ncRNA = []
                snps_in_pseudogene = []
                intergenic = []
                for eachAccession in gbank_accessions:
                    if eachAccession.split(':')[0] in eachKEY or eachKEY in eachAccession:

                        annotationDB = GenbankContent[eachAccession]
                        for eachSNP in self.filteredSNPs[eachKEY]:
             
                            flag = False
                            posSNP = eachSNP[0]
                            total_depth = eachSNP[3]
                            nb_reads_reference = eachSNP[4]
                            nb_reads_mutated = eachSNP[5]
                            ratio = eachSNP[6]
                            quality = eachSNP[7]

                            for eachElem in annotationDB:
                                if posSNP <= int(eachElem[3]) and posSNP >= int(eachElem[2]):
                                    flag = True
                                    if eachElem[1] == 'CDS':
                                        codons_nuc_sequence = list(zip(range(1, len(eachElem[6]), 3), [
                                            eachElem[6][i:i+3] for i in range(0, len(eachElem[6]), 3)]))
                                        gene_name = eachElem[0]
                                        gene_fucntion = eachElem[5]
                                        gene_orientation = eachElem[4]
                                        if gene_orientation == '+':
                                            pos_snp_in_gene = (
                                                eachSNP[0] - int(eachElem[2])) + 1
                                            reference_allele = eachSNP[1]
                                            mutated_allele = eachSNP[2]
                                        #TODO:even if the gene is in the reverse strand i decided to keep the SNPs as according to genome. in the other hand, it will be better to change the codon as follorw to show the position of the SNP C(G)T or (C)GT or CG(T)
                                        else:
                                            pos_snp_in_gene = (
                                                int(eachElem[3]) - eachSNP[0]) + 1
                                            reference_allele = eachSNP[1]
                                            mutated_allele = eachSNP[2]

                                        for i in range(len(codons_nuc_sequence)):
                                            if codons_nuc_sequence[i][0] <= pos_snp_in_gene < codons_nuc_sequence[i][0] + 3:
                                                ref_codon = codons_nuc_sequence[i][1]
                                                position_nuc_ToBe_changed = (
                                                    pos_snp_in_gene - codons_nuc_sequence[i][0])
                                                if gene_orientation == '+':
                                                    new_codon = ref_codon[:position_nuc_ToBe_changed] + \
                                                        '['+mutated_allele + ']'+\
                                                        ref_codon[position_nuc_ToBe_changed+1:]
                                                    new_codon_x = ref_codon[:position_nuc_ToBe_changed] + \
                                                        mutated_allele +\
                                                        ref_codon[position_nuc_ToBe_changed+1:]
                                                    ref_AA = standcode[ref_codon.lower(
                                                    )]
                                                    new_AA = standcode[new_codon_x.lower(
                                                    )]
                                                else:
                                                    new_codon = ref_codon[:position_nuc_ToBe_changed] + \
                                                        '['+Reverse_nuc[mutated_allele] + ']'+\
                                                        ref_codon[position_nuc_ToBe_changed+1:]
                                                    ref_AA = standcode[ref_codon.lower(
                                                    )]
                                                    new_codon_x = ref_codon[:position_nuc_ToBe_changed] + \
                                                        Reverse_nuc[mutated_allele] + \
                                                        ref_codon[position_nuc_ToBe_changed+1:]                                                    
                                                    new_AA = standcode[new_codon_x.lower(
                                                    )]
                                                collected_data = [posSNP, reference_allele, mutated_allele, total_depth, nb_reads_reference, nb_reads_mutated, ratio,
                                                                  quality, gene_name, gene_fucntion, gene_orientation, pos_snp_in_gene, ref_codon, new_codon, ref_AA, new_AA, i+1]
                                                if ref_AA == new_AA:
                                                    collected_data.append('Syn')

                                                else:
                                                    collected_data.append('NS')
                                                snps_in_genes.append(
                                                    collected_data)

                                    elif eachElem[1] == 'RBS':
                                        snps_in_RBS.append(
                                            eachSNP + ['RBS'] + ['.'] * 9)

                                    elif eachElem[1] == 'tRNA':
                                        snps_in_tRNA.append(
                                            eachSNP + ['tRNA'] + ['.'] * 9)

                                    elif eachElem[1] == 'rRNA':
                                        snps_in_rRNA.append(
                                            eachSNP + ['rRNA'] + ['.'] * 9)

                                    elif eachElem[1] == 'ncRNA':
                                        snps_in_ncRNA.append(
                                            eachSNP + ['ncRNA'] + ['.'] * 9)

                                    elif eachElem[1] == 'pseudogene':
                                        codons_nuc_sequence = list(zip(range(1, len(eachElem[6]), 3), [
                                            eachElem[6][i:i+3] for i in range(0, len(eachElem[6]), 3)]))
                                        gene_name = eachElem[0]+'-pseudogene'
                                        gene_fucntion = eachElem[5]
                                        gene_orientation = eachElem[4]
                                        if gene_orientation == '+':
                                            pos_snp_in_gene = (
                                                eachSNP[0] - int(eachElem[2])) + 1
                                            reference_allele = eachSNP[1]
                                            mutated_allele = eachSNP[2]
                                        #TODO:even if the gene is in the reverse strand i decided to keep the SNPs as according to genome. in the other hand, it will be better to change the codon as follorw to show the position of the SNP C(G)T or (C)GT or CG(T)
                                        else:
                                            pos_snp_in_gene = (
                                                int(eachElem[3]) - eachSNP[0]) + 1
                                            reference_allele = eachSNP[1]
                                            mutated_allele = eachSNP[2]

                                        for i in range(len(codons_nuc_sequence)):
                                            if codons_nuc_sequence[i][0] <= pos_snp_in_gene < codons_nuc_sequence[i][0] + 3:
                                                ref_codon = codons_nuc_sequence[i][1]
                                                position_nuc_ToBe_changed = (
                                                    pos_snp_in_gene - codons_nuc_sequence[i][0])
                                                if gene_orientation == '+':
                                                    new_codon = ref_codon[:position_nuc_ToBe_changed] + \
                                                        '['+mutated_allele + ']'+\
                                                        ref_codon[position_nuc_ToBe_changed+1:]
                                                    new_codon_x = ref_codon[:position_nuc_ToBe_changed] + \
                                                        mutated_allele +\
                                                        ref_codon[position_nuc_ToBe_changed+1:]
                                                    ref_AA = standcode[ref_codon.lower(
                                                    )]
                                                    new_AA = standcode[new_codon_x.lower(
                                                    )]
                                                else:
                                                    new_codon = ref_codon[:position_nuc_ToBe_changed] + \
                                                        '['+Reverse_nuc[mutated_allele] + ']'+\
                                                        ref_codon[position_nuc_ToBe_changed+1:]
                                                    ref_AA = standcode[ref_codon.lower(
                                                    )]
                                                    new_codon_x = ref_codon[:position_nuc_ToBe_changed] + \
                                                        Reverse_nuc[mutated_allele] + \
                                                        ref_codon[position_nuc_ToBe_changed+1:]                                                    
                                                    new_AA = standcode[new_codon_x.lower(
                                                    )]
                                                collected_data = [posSNP, reference_allele, mutated_allele, total_depth, nb_reads_reference, nb_reads_mutated, ratio,
                                                                  quality, gene_name, gene_fucntion, gene_orientation, pos_snp_in_gene, ref_codon, new_codon, ref_AA, new_AA, i+1]
                                                if ref_AA == new_AA:
                                                    collected_data.append('Syn')

                                                else:
                                                    collected_data.append('NS')
                                                snps_in_genes.append(
                                                    collected_data)

                                        snps_in_pseudogene.append(eachSNP)
                            if flag == False:

                                cutoff_distance = 100
                                if posSNP < int(annotationDB[0][2]):
                                    if annotationDB[0][4] == '+':
                                        distance_gene = int(
                                            annotationDB[0][2]) - posSNP
                                        if distance_gene <= cutoff_distance:
                                            intergenic.append(eachSNP + ['intergenic', '.', '+', str(
                                                distance_gene) + ' bp from ' + annotationDB[0][0]] + ['.'] * 6)
                                        else:
                                            intergenic.append(
                                                eachSNP + ['intergenic', '.', '+', '.'] + ['-'] * 6)

                                    else:
                                        intergenic.append(
                                            eachSNP + ['intergenic', '.', '+', '.'] + ['-'] * 6)
                                else:
                                    if posSNP > int(annotationDB[-1][3]):
                                        if annotationDB[-1][4] == '-':
                                            distance_gene = posSNP -\
                                                int(annotationDB[0][3])
                                            if distance_gene <= cutoff_distance:
                                                intergenic.append(eachSNP + ['intergenic', '.', '+', str(
                                                    distance_gene) + ' bp from ' + annotationDB[-1][0]] + ['.'] * 6)
                                            else:
                                                intergenic.append(
                                                    eachSNP + ['intergenic', '.', '+', '.'] + ['-'] * 6)

                                        else:
                                            intergenic.append(
                                                eachSNP + ['intergenic', '.', '+', '.'] + ['-'] * 6)

                                k = 0
                                while k < len(annotationDB) - 1:
                                    if int(annotationDB[k][3]) < posSNP < int(annotationDB[k + 1][2]):
                                        if annotationDB[k][4] == '+' and annotationDB[k + 1][4] == '+':
                                            distance_gene = int(
                                                annotationDB[k + 1][2]) - posSNP
                                            if distance_gene <= cutoff_distance:
                                                intergenic.append(eachSNP + ['intergenic', '.', '+', str(
                                                    distance_gene) + ' bp from ' + annotationDB[k + 1][0]] + ['.'] * 6)
                                            else:
                                                intergenic.append(
                                                    eachSNP + ['intergenic', '.', '+', '.'] + ['.'] * 6)

                                        elif annotationDB[k][4] == '-' and annotationDB[k + 1][4] == '+':
                                            distance_gene1 = posSNP - \
                                                int(annotationDB[k][3])
                                            distance_gene2 = int(
                                                annotationDB[k + 1][2]) - posSNP
                                            if distance_gene1 <= cutoff_distance and distance_gene2 <= cutoff_distance:
                                                intergenic.append(eachSNP + ['intergenic', '.', '+', str(distance_gene1) + ' bp from ' + annotationDB[k][0] + ' and ' + str(
                                                    distance_gene2) + ' bp from ' + annotationDB[k + 1][0]] + ['.'] * 6)
                                            elif distance_gene1 <= cutoff_distance and distance_gene2 > cutoff_distance:
                                                intergenic.append(eachSNP + ['intergenic', '.', '+', str(
                                                    distance_gene1) + ' bp from ' + annotationDB[k][0]] + ['.'] * 6)
                                            elif distance_gene2 <= cutoff_distance and distance_gene1 > cutoff_distance:
                                                intergenic.append(eachSNP + ['intergenic', '.', '+', str(
                                                    distance_gene2) + ' bp from ' + annotationDB[k + 1][0]] + ['.'] * 6)
                                            else:
                                                intergenic.append(
                                                    eachSNP + ['intergenic', '.', '+', '.'] + ['-'] * 6)

                                        elif annotationDB[k][4] == '-' and annotationDB[k + 1][4] == '-':
                                            distance_gene = posSNP - \
                                                int(annotationDB[k][3])
                                            if distance_gene <= cutoff_distance:
                                                intergenic.append(eachSNP + ['intergenic', '.', '+', str(
                                                    distance_gene) + ' bp from ' + annotationDB[k][0]] + ['.'] * 6)
                                            else:
                                                intergenic.append(
                                                    eachSNP + ['intergenic', '.', '+', '.'] + ['.'] * 6)

                                        else:

                                            intergenic.append(
                                                eachSNP + ['intergenic', '.', '+', '.'] + ['.'] * 6)

                                    k += 1

                        data_collector[eachAccession] = {'Genes': snps_in_genes, 'tRNA': snps_in_tRNA, 'RBS': snps_in_RBS,
                                                         'rRNA': snps_in_rRNA, 'ncRNA': snps_in_ncRNA, 'Pseudogenes': snps_in_pseudogene, 'intergenic': intergenic}

        return data_collector



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='snptoolkit',
      version='2.2.4',
      description='SNPs filtering, annotation and combining',
      url='https://github.com/Amine-Namouchi/snpToolkit',
      author='Amine Namouchi',
      author_email='amine.namouchi@gmail.com',
      python_requires='>=3.0',
      long_description='SNPs annotation, filtering and combining from vcf files',
      license='GPLv3+',
      classifiers=[
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
          'Programming Language :: Python :: 3',
          'Operating System :: OS Independent',
      ],
      keywords='SNPs filtering annotation variants calling',
      install_requires=['pysam', 'pandas', 'tqdm', 'coloredlogs', 'Biopython','dash','plotly'],
      scripts=['snptoolkit','./snpToolkit_modules/annotate_snpToolkit.py','./snpToolkit_modules/annotate_snpToolkit.py','./snpToolkit_modules/calls_snpToolkit.py', './snpToolkit_modules/argsLogger_snpToolkit.py','./snpToolkit_modules/combine_snpToolkit.py','./snpToolkit_modules/expand_snpToolkit.py','./snpToolkit_modules/explore_snpToolkit.py','./snpToolkit_modules/plot_snptoolkit_output.py']
      )
